#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import logging

import pytest

logging.basicConfig(level=logging.DEBUG)

from foodx_devops_tools.deploy_me._status import (
    DeploymentState,
    DeploymentStatus,
    all_completed,
    all_success,
    any_completed_dirty,
)

TIMEOUT_SECONDS = 10


class TestAllSuccess:
    def test_clean(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.success),
        ]
        result = all_success(mock_input)
        assert result

    def test_not_clean1(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.failed),
        ]
        result = all_success(mock_input)
        assert not result

    def test_not_clean2(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.cancelled),
        ]
        result = all_success(mock_input)
        assert not result

    def test_not_clean3(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.pending),
        ]
        result = all_success(mock_input)
        assert not result

    def test_not_clean4(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.in_progress),
        ]
        result = all_success(mock_input)
        assert not result


class TestAllCompleted:
    def test_clean1(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.failed),
            DeploymentState(DeploymentState.ResultType.cancelled),
        ]
        result = all_completed(mock_input)
        assert result

    def test_clean2(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.success),
        ]
        result = all_completed(mock_input)
        assert result

    def test_clean3(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.failed),
            DeploymentState(DeploymentState.ResultType.failed),
            DeploymentState(DeploymentState.ResultType.failed),
        ]
        result = all_completed(mock_input)
        assert result

    def test_clean4(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.cancelled),
            DeploymentState(DeploymentState.ResultType.cancelled),
            DeploymentState(DeploymentState.ResultType.cancelled),
        ]
        result = all_completed(mock_input)
        assert result

    def test_not_clean1(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.in_progress),
        ]
        result = all_completed(mock_input)
        assert not result

    def test_not_clean2(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.pending),
        ]
        result = all_completed(mock_input)
        assert not result


class TestAnyCompletedDirty:
    def test_clean1(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.failed),
        ]
        result = any_completed_dirty(mock_input)
        assert result

    def test_clean2(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.cancelled),
        ]
        result = any_completed_dirty(mock_input)
        assert result

    def test_in_progress(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.failed),
            DeploymentState(DeploymentState.ResultType.in_progress),
        ]
        result = any_completed_dirty(mock_input)
        assert result

    def test_pending(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.cancelled),
            DeploymentState(DeploymentState.ResultType.pending),
        ]
        result = any_completed_dirty(mock_input)
        assert result

    def test_all_success(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.success),
        ]
        result = any_completed_dirty(mock_input)
        assert not result

    def test_in_progress(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.in_progress),
        ]
        result = any_completed_dirty(mock_input)
        assert not result

    def test_pending(self):
        mock_input = [
            DeploymentState(DeploymentState.ResultType.success),
            DeploymentState(DeploymentState.ResultType.pending),
        ]
        result = any_completed_dirty(mock_input)
        assert not result


@pytest.fixture()
def status_instance():
    def _apply(timeout_seconds: float = 10):
        _status_instance = DeploymentStatus("some.context", timeout_seconds)

        return _status_instance

    return _apply


@pytest.fixture()
def simple_status(status_instance):
    async def _apply(timeout_seconds: float):
        under_test = status_instance(timeout_seconds=timeout_seconds)
        await under_test.initialize("n1")
        await under_test.initialize("n2")

        under_test.start_monitor()

        return under_test

    return _apply


class TestDeploymentStatus:
    @pytest.mark.asyncio
    async def test_clean(self, status_instance):
        under_test = status_instance()
        this_name = "some_name"

        try:
            assert not await under_test.names()
            await asyncio.wait_for(
                under_test.initialize(this_name), timeout=TIMEOUT_SECONDS
            )
            under_test.start_monitor()

            assert await asyncio.wait_for(
                under_test.names(), timeout=TIMEOUT_SECONDS
            ) == {this_name}

            await under_test.write(
                this_name,
                DeploymentState.ResultType.pending,
                message="some message",
            )

            # pause slightly to enable queue to be processed
            await asyncio.sleep(0.1)

            result = await asyncio.wait_for(
                under_test.read(this_name), timeout=TIMEOUT_SECONDS
            )

            assert result.code == DeploymentState.ResultType.pending
            assert result.message == "some message"
        except asyncio.TimeoutError:
            pytest.fail("status test timed out")


class TestAllCompletedEvent:
    @pytest.mark.asyncio
    async def test_success(self, simple_status):
        under_test = await simple_status(timeout_seconds=5)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n1", DeploymentState.ResultType.success)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.success)

        waiter_task = asyncio.create_task(under_test.wait_for_all_completed())
        # schedule the status changes to be detected
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending

        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending

        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.success

    @pytest.mark.asyncio
    async def test_success_fail(self, simple_status):
        under_test = await simple_status(timeout_seconds=2)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n1", DeploymentState.ResultType.success)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.failed)

        waiter_task = asyncio.create_task(under_test.wait_for_all_completed())
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.failed

    @pytest.mark.asyncio
    async def test_success_cancelled(self, simple_status):
        under_test = await simple_status(timeout_seconds=2)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n1", DeploymentState.ResultType.success)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.cancelled)

        waiter_task = asyncio.create_task(under_test.wait_for_all_completed())
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.cancelled

    @pytest.mark.asyncio
    async def test_in_progress(self, simple_status):
        under_test = await simple_status(timeout_seconds=1)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)

        waiter_task = asyncio.create_task(under_test.wait_for_all_completed())
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        with pytest.raises(asyncio.TimeoutError):
            await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

    @pytest.mark.asyncio
    async def test_timeout_raises(self, simple_status):
        under_test = await simple_status(timeout_seconds=0.5)

        async def mock_status_updated():
            # do nothing so that state never transitions to a completed state.
            pass

        waiter_task = asyncio.create_task(under_test.wait_for_all_completed())
        asyncio.create_task(mock_status_updated())

        with pytest.raises(asyncio.TimeoutError):
            await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending

    @pytest.mark.asyncio
    async def test_prior_success(self, simple_status):
        """completion already flagged on entry is clean."""
        under_test = await simple_status(timeout_seconds=5)

        # both these deployments have completed (failed, cancelled)
        await under_test.write("n1", DeploymentState.ResultType.failed)
        await under_test.write("n2", DeploymentState.ResultType.cancelled)

        # need an await here to allow other tasks to execute (especially
        # processing the queue)
        await asyncio.sleep(0.01)

        waiter_task = asyncio.create_task(under_test.wait_for_all_completed())

        # status should now consistently read as those states
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.failed
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.cancelled

        await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.failed
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.cancelled


class TestAllSucceededEvent:
    @pytest.mark.asyncio
    async def test_success(self, simple_status):
        under_test = await simple_status(timeout_seconds=5)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n1", DeploymentState.ResultType.success)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.success)

        waiter_task = asyncio.create_task(under_test.wait_for_all_succeeded())
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.success

    @pytest.mark.asyncio
    async def test_success_fail(self, simple_status):
        under_test = await simple_status(timeout_seconds=2)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n1", DeploymentState.ResultType.success)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.failed)

        waiter_task = asyncio.create_task(under_test.wait_for_all_succeeded())
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        with pytest.raises(asyncio.TimeoutError):
            await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.failed

    @pytest.mark.asyncio
    async def test_success_cancelled(self, simple_status):
        under_test = await simple_status(timeout_seconds=2)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n1", DeploymentState.ResultType.success)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.cancelled)

        waiter_task = asyncio.create_task(under_test.wait_for_all_succeeded())
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        with pytest.raises(asyncio.TimeoutError):
            await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.cancelled

    @pytest.mark.asyncio
    async def test_in_progress(self, simple_status):
        under_test = await simple_status(timeout_seconds=1)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)

        waiter_task = asyncio.create_task(under_test.wait_for_all_succeeded())
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        with pytest.raises(asyncio.TimeoutError):
            await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

    @pytest.mark.asyncio
    async def test_timeout_raises(self, simple_status):
        under_test = await simple_status(timeout_seconds=0.5)

        async def mock_status_updated():
            # do nothing so that state never transitions to a completed state.
            pass

        waiter_task = asyncio.create_task(under_test.wait_for_all_succeeded())
        asyncio.create_task(mock_status_updated())

        with pytest.raises(asyncio.TimeoutError):
            await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending

    @pytest.mark.asyncio
    async def test_prior_success(self, simple_status):
        """success already flagged on entry is clean."""
        under_test = await simple_status(timeout_seconds=5)

        await under_test.write("n1", DeploymentState.ResultType.success)
        await under_test.write("n2", DeploymentState.ResultType.success)

        # need an await here to allow other tasks to execute (especially
        # processing the queue)
        await asyncio.sleep(0.01)
        waiter_task = asyncio.create_task(under_test.wait_for_all_succeeded())

        # the status will show as the initial state until the queue of state
        # updates is cleared.
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.success

        await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.success
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.success


class TestCompletedNamedEvent:
    @pytest.mark.asyncio
    async def test_success(self, simple_status):
        under_test = await simple_status(timeout_seconds=5)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n1", DeploymentState.ResultType.failed)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.success)

        waiter_task = asyncio.create_task(under_test.wait_for_completion("n2"))
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n2")
        ).code != DeploymentState.ResultType.success

        await waiter_task

        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.success

    @pytest.mark.asyncio
    async def test_success_fail(self, simple_status):
        under_test = await simple_status(timeout_seconds=2)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n1", DeploymentState.ResultType.success)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.failed)

        waiter_task = asyncio.create_task(under_test.wait_for_completion("n2"))
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n2")
        ).code != DeploymentState.ResultType.failed

        await waiter_task

        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.failed

    @pytest.mark.asyncio
    async def test_in_progress(self, simple_status):
        under_test = await simple_status(timeout_seconds=1)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n2", DeploymentState.ResultType.in_progress)

        waiter_task = asyncio.create_task(under_test.wait_for_completion("n2"))
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending
        await asyncio.sleep(0.2)
        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

        with pytest.raises(asyncio.TimeoutError):
            await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.in_progress
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.in_progress

    @pytest.mark.asyncio
    async def test_timeout_raises(self, simple_status):
        under_test = await simple_status(timeout_seconds=0.5)

        async def mock_status_updated():
            # do nothing so that state never transitions to a completed state.
            pass

        waiter_task = asyncio.create_task(under_test.wait_for_completion("n2"))
        asyncio.create_task(mock_status_updated())

        with pytest.raises(asyncio.TimeoutError):
            await waiter_task

        assert (
            await under_test.read("n1")
        ).code == DeploymentState.ResultType.pending
        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.pending

    @pytest.mark.asyncio
    async def test_prior_success(self, simple_status):
        """success already flagged on entry is clean."""
        under_test = await simple_status(timeout_seconds=5)

        async def mock_status_updated():
            nonlocal under_test

            await asyncio.sleep(0.1)
            await under_test.write("n1", DeploymentState.ResultType.in_progress)
            await asyncio.sleep(0.2)
            await under_test.write("n1", DeploymentState.ResultType.failed)

        await under_test.write("n2", DeploymentState.ResultType.success)

        # need an await here to allow other tasks to execute (especially
        # processing the queue)
        await asyncio.sleep(0.01)

        waiter_task = asyncio.create_task(under_test.wait_for_completion("n2"))
        asyncio.create_task(mock_status_updated())

        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.success

        await waiter_task

        assert (
            await under_test.read("n2")
        ).code == DeploymentState.ResultType.success



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import pytest

from foodx_devops_tools.puff.arm import _do_file_actions


class TestDoFileActions:
    MOCK_PARAMETER_DATA = {
        "some-file": {
            "k1": "v1",
        }
    }

    @pytest.mark.asyncio
    async def test_delete_clean(self, mocker):
        mock_delete = mocker.patch(
            "foodx_devops_tools.puff.arm._delete_parameter_file"
        )
        mocker.patch("foodx_devops_tools.puff.arm._save_parameter_file")

        await _do_file_actions(
            True, pathlib.Path("some/path"), self.MOCK_PARAMETER_DATA, False
        )

        mock_delete.assert_called_once()

    @pytest.mark.asyncio
    async def test_create_clean(self, mocker):
        mocker.patch("foodx_devops_tools.puff.arm._delete_parameter_file")
        mock_save = mocker.patch(
            "foodx_devops_tools.puff.arm._save_parameter_file"
        )

        await _do_file_actions(
            False, pathlib.Path("some/path"), self.MOCK_PARAMETER_DATA, False
        )

        mock_save.assert_called_once()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Deployment status monitoring."""

import asyncio
import copy
import logging
import typing

import click

from foodx_devops_tools.pipeline_config import (
    DependencyDeclarations,
    IterationContext,
)

from ._exceptions import DeploymentTerminatedError
from ._status import (
    DeploymentState,
    DeploymentStatus,
    all_success,
    any_completed_dirty,
)

log = logging.getLogger(__name__)

STATUS_KEY_MAX_RETRIES = 20
STATUS_KEY_RETRY_SLEEP_SECONDS = 10


def _generate_dependency_contexts(
    this_context: IterationContext, dependency_names: typing.Set[str]
) -> typing.Set[str]:
    dependency_context = set()
    for x in dependency_names:
        base = copy.deepcopy(this_context)
        base[-1] = x
        dependency_context.add(str(base))

    return dependency_context


async def _confirm_dependency_entity_status(
    dependency_names: typing.Set[str],
    entity_status: DeploymentStatus,
    this_context: IterationContext,
    status_monitor_sleep_seconds: float,
) -> typing.Set[str]:
    """
    Check that dependency entities are present in entity status.

    Entities here are either frames or applications.
    """

    async def report_missing_dependencies(
        this_state: DeploymentState.ResultType,
        console_report: bool,
    ) -> str:
        nonlocal dependency_names, entity_status, this_context

        message = "dependency not in entity status, {0}, {1}".format(
            this_context, str(dependency_names)
        )
        log.warning(message)
        if console_report:
            click.echo(click.style(message, fg="yellow"))
        await entity_status.write(
            str(this_context),
            this_state,
            message=message,
        )

        return message

    dependency_contexts = _generate_dependency_contexts(
        this_context, dependency_names
    )

    # we know the key is *supposed* to exist because it's defined in
    # `depends_on` (although it could be a typo), so keep trying in
    # case we just polled too early and it hasn't been registered
    # yet. the wait timeout will take care of it being the wrong value.
    not_found = True
    attempt_count = 0
    while not_found and (attempt_count < STATUS_KEY_MAX_RETRIES):
        entity_contexts = await entity_status.names()
        if all([x in entity_contexts for x in dependency_contexts]):
            not_found = False
            log.debug(
                "all dependencies found in status, {0}, {1}, "
                "({2} retries)".format(
                    entity_contexts, dependency_contexts, attempt_count
                )
            )
        else:
            attempt_count += 1
            await report_missing_dependencies(
                DeploymentState.ResultType.pending, False
            )
            await asyncio.sleep(status_monitor_sleep_seconds)

    if not_found:
        message = await report_missing_dependencies(
            DeploymentState.ResultType.failed, True
        )
        raise DeploymentTerminatedError(message)

    return dependency_contexts


async def wait_for_dependencies(
    iteration_context: IterationContext,
    dependency_data: DependencyDeclarations,
    entity_status: DeploymentStatus,
) -> None:
    """
    Wait for dependency entity status to succeed or fail.

    A timeout occurs if the specified duration is exceeded.

    Args:
        iteration_context: Current deployment hierarchy object.
        dependency_data: Entity dependency data.
        entity_status: Status reporting object for entities.

    Raises:
        DeploymentTerminatedError: If any dependencies don't complete or the
                                  maximum wait duration is exceeded.
    """

    async def cancel_deployment() -> None:
        nonlocal dependency_data, entity_status, iteration_context

        message = "cancelled due to dependency failure, {0}, {1}".format(
            iteration_context, str(dependency_data)
        )
        log.error(message)
        click.echo(click.style(message, fg="red"))
        await entity_status.write(
            str(iteration_context),
            DeploymentState.ResultType.cancelled,
            message=message,
        )
        raise DeploymentTerminatedError(message)

    async def report_success() -> None:
        nonlocal entity_status, iteration_context

        message = (
            "dependencies completed. proceeding with deployment, {0}".format(
                iteration_context
            )
        )
        log.info(message)
        click.echo(click.style(message, fg="cyan"))
        await entity_status.write(
            str(iteration_context), DeploymentState.ResultType.in_progress
        )

    # if there are no dependencies just skip dependency processing.
    if dependency_data:
        dependency_names = set(dependency_data)
        dependency_contexts = await _confirm_dependency_entity_status(
            dependency_names,
            entity_status,
            iteration_context,
            STATUS_KEY_RETRY_SLEEP_SECONDS,
        )

        message = "waiting for dependency completion, {0}, {1}".format(
            iteration_context, dependency_contexts
        )
        log.info(message)
        click.echo(click.style(message))
        try:
            await asyncio.gather(
                *[
                    entity_status.wait_for_completion(x)
                    for x in dependency_contexts
                ]
            )

            dependency_status = [
                await entity_status.read(x) for x in dependency_contexts
            ]
            if all_success(dependency_status):
                await report_success()
            elif any_completed_dirty(dependency_status):
                await cancel_deployment()
        except asyncio.TimeoutError:
            await entity_status.write(
                str(iteration_context),
                DeploymentState.ResultType.cancelled,
                message=message,
            )
            message = (
                "deployment cancelled due to timeout waiting for "
                "dependency completion, {0}".format(str(iteration_context))
            )
            log.error(message)
            click.echo(click.style(message, fg="red"))
            raise DeploymentTerminatedError(message)
    else:
        log.debug("Skipping empty dependencies for status")



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""File I/O for pipeline configuration metadata."""

from ._checks import do_path_check  # noqa: F401
from ._paths import PipelineConfigurationPaths  # noqa: F401
from .clients import ClientsDefinition, load_clients  # noqa: F401
from .deployments import DeploymentsDefinition, load_deployments  # noqa: F401
from .frames import (  # noqa: F401
    ApplicationDefinition,
    ApplicationDeploymentSteps,
    DependencyDeclarations,
    FramesDefinition,
    SingularFrameDefinition,
    StructuredName,
    StructuredPathCollection,
    load_frames,
)
from .pipeline import PipelineConfiguration  # noqa: F401
from .release_states import (  # noqa: F401
    ReleaseStatesDefinition,
    load_release_states,
)
from .service_principals import (  # noqa: F401
    ServicePrincipals,
    load_service_principals,
)
from .static_secrets import StaticSecrets, load_static_secrets  # noqa: F401
from .subscriptions import (  # noqa: F401
    SubscriptionsDefinition,
    load_subscriptions,
)
from .systems import SystemsDefinition, load_systems  # noqa: F401
from .template_context import (  # noqa: F401
    TemplateContext,
    load_template_context,
)
from .tenants import TenantsDefinition, load_tenants  # noqa: F401
from .views import (  # noqa: F401
    DeploymentContext,
    FlattenedDeployment,
    IterationContext,
    ReleaseView,
)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""General support for templating."""

import asyncio
import dataclasses
import logging
import pathlib

import pydantic

from foodx_devops_tools.puff import run_puff
from foodx_devops_tools.utilities.jinja2 import (
    FrameTemplates,
    TemplateParameters,
)

from ._exceptions import TemplateError

log = logging.getLogger(__name__)

JINJA_FILE_PREFIX = "jinja2."


class TemplatePaths(pydantic.BaseModel):
    """Collection of file paths for ARM templates."""

    source: pathlib.Path
    target: pathlib.Path


@dataclasses.dataclass
class TemplateFiles:
    """Collection of file paths for template processing."""

    arm_template: TemplatePaths
    arm_template_parameters: TemplatePaths


@dataclasses.dataclass
class ArmTemplateDeploymentFiles:
    """Paths to the two files needed for ARM template deployment."""

    arm_template: pathlib.Path
    parameters: pathlib.Path


def json_inlining(content: str) -> str:
    """
    Inline text content for inclusion in a JSON file (ARM template).

    Escape double quotes and retain newlines as escapes to embody the content
    in a single line.

    Args:
        content: Content to be "inlined".

    Returns:
        Modified content.
    """
    escaped_newlines = repr(content).strip("'").replace(r"\n", r"\n")
    escaped_quotes = escaped_newlines.replace('"', r"\"")

    return escaped_quotes


async def _apply_template(
    template_environment: FrameTemplates,
    source_file: pathlib.Path,
    target_file: pathlib.Path,
    parameters: TemplateParameters,
) -> None:
    """
    Apply frame-specific template and parameters ready for deployment.

    Args:
        template_environment:   Frame-specific template environment.
        source_file:            Source template file.
        target_directory:       Target directory in which to store fulfilled
                                template file.
        parameters:             Parameter to apply to the template file.

    Returns:
        Target file path of the fulfilled template.
    """
    log.debug(
        "Applying jinja2 templating, {0} (source), "
        "{1} (destination)".format(source_file, target_file)
    )
    await template_environment.apply_template(
        source_file.name, target_file, parameters
    )


def _verify_puff_target(file_path: pathlib.Path) -> None:
    if not file_path.is_file():
        message = (
            f"Expected puff generated ARM template parameter file is "
            f"missing, {file_path}"
        )
        log.error(message)
        raise TemplateError(message)


async def _prepare_working_directory(working_dir: pathlib.Path) -> None:
    """
    Ensure that the working directory exists.

    Args:
        working_dir: Expected path of working directory.

    Raises:
        TemplateError:  If the working directory path exists, but is not
                        actually a directory.
    """
    if not working_dir.exists():
        working_dir.mkdir(parents=True, exist_ok=True)
    elif not working_dir.is_dir():
        raise TemplateError(
            f"working directory name exists but is not a directory, "
            f"{working_dir}"
        )


def _log_arm_template_paths(arm_template: TemplatePaths) -> None:
    log.debug(f"source_arm_template_path, {arm_template.source}")
    log.debug(f"arm_target_file, {arm_template.target}")

    log.info(f"applying jinja2 to ARM template file, {arm_template.source}")


def _construct_arm_template_parameter_paths(
    arm_template_parameters: TemplatePaths,
) -> pathlib.Path:
    source_puff_file_path = arm_template_parameters.source

    # puffd_parameters_target_file: the expected arm template parameter file
    # generated by puff.
    puffd_parameters_target_file = arm_template_parameters.target
    log.debug(
        f"puffd_parameters_target_file json, {puffd_parameters_target_file}"
    )
    parameters_target_dir = puffd_parameters_target_file.parent
    log.debug(f"parameters_target_dir, {parameters_target_dir}")

    # jinjad_parameters_target_file: the expected arm template parameter file
    # generated by jinja2 template processing.
    jinjad_parameters_target_file = (
        parameters_target_dir / f"jinjad.{puffd_parameters_target_file.name}"
    )
    log.debug(
        f"jinjad_parameters_target_file json, {jinjad_parameters_target_file}"
    )

    log.info(
        f"applying jinja2 to ARM template parameter file,"
        f" {source_puff_file_path}"
    )

    return jinjad_parameters_target_file


async def prepare_deployment_files(
    template_files: TemplateFiles,
    parameters: TemplateParameters,
) -> ArmTemplateDeploymentFiles:
    """
    Prepare final ARM template and parameter files for deployment.

    Args:
        template_files: Paths to source files for processing.
        parameters: Parameters to be applied to templates.

    Returns:
        Paths to ARM template and ARM template parameter files.
    Raises:
        TemplateError:  If an error occurs during puff or template processing.
    """
    arm_source = template_files.arm_template.source
    arm_target = template_files.arm_template.target
    _log_arm_template_paths(template_files.arm_template)

    # the puff YAML file.
    parameters_source = template_files.arm_template_parameters.source
    # the arm template parameters file generated from the puff run.
    puffd_parameters_target = template_files.arm_template_parameters.target
    # the arm template parameter file generated by jinja2 processing.
    parameters_target = _construct_arm_template_parameter_paths(
        template_files.arm_template_parameters
    )

    # folders containing _jinja template_ source files
    template_paths = (
        [arm_source.parent, parameters_target.parent]
        if arm_source.parent != parameters_target.parent
        else [arm_source.parent]
    )
    log.debug(f"frame template paths, {template_paths}")
    template_environment = FrameTemplates(template_paths)
    template_environment.environment.filters["json_inlining"] = json_inlining

    await _prepare_working_directory(parameters_target.parent)
    if parameters_target.parent != arm_target.parent:
        # also prepare the distinct arm target directory
        await _prepare_working_directory(arm_target.parent)

    # transform the puff file to arm template parameter json files.
    await run_puff(
        parameters_source,
        False,
        False,
        disable_ascii_art=True,
        output_dir=puffd_parameters_target.parent,
    )
    _verify_puff_target(puffd_parameters_target)

    # now process jinja2 templates against JSON files.
    await asyncio.gather(
        _apply_template(
            template_environment,
            puffd_parameters_target,
            parameters_target,
            parameters,
        ),
        _apply_template(
            template_environment,
            arm_source,
            arm_target,
            parameters,
        ),
    )
    result = ArmTemplateDeploymentFiles(
        arm_template=arm_target,
        parameters=parameters_target,
    )
    return result



#  Copyright (c) 2022 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import logging

import click

log = logging.getLogger(__name__)


async def delay_step(delay_seconds: float) -> None:
    """
    Pause a deployment for the specified duration.

    Args:
        delay_seconds: Pause delay in seconds.
    """
    message = "application steps paused for, {0} (seconds)".format(
        delay_seconds
    )
    log.info(message)
    click.echo(message)
    await asyncio.sleep(delay_seconds)



#  Copyright (c) 2022 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.deploy_me.application_steps._deploy import (
    _construct_resource_group_name,
)


class TestConstructResourceGroupName:
    def test_clean_user_precedence(self):
        result = _construct_resource_group_name("client", "some_name")

        assert result == "client-some_name"



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['foodx_devops_tools',
 'foodx_devops_tools.azure',
 'foodx_devops_tools.azure.cloud',
 'foodx_devops_tools.deploy_me',
 'foodx_devops_tools.deploy_me.application_steps',
 'foodx_devops_tools.pipeline_config',
 'foodx_devops_tools.puff',
 'foodx_devops_tools.release_flow',
 'foodx_devops_tools.utilities']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles >=0.7.0, <1.0',
 'ansible >=4.4.0, <6.0',
 'azure-devops >=6.0.0b4, <7.0',
 'click >=8.1.2, <9',
 'deepmerge >=1.0.1, <2',
 'jinja2 >=3.1.1, <4',
 'pydantic >=1.9.0, <2',
 'ruamel.yaml >=0.17.20, <1.0']

extras_require = \
{'dev': ['build_harness >=1.0, <2',
         'pre_commit >=2.17.0, <3',
         'pyjson5 >=1.5.2, <2',
         'types-aiofiles >=0.1.9, <1.0'],
 'doc': ['sphinx >=4.5.0, <5', 'sphinx_rtd_theme >=1.0, <2'],
 'test': ['asynctest >=0.13.0, <1.0',
          'pytest >=7.1.1, <8',
          'pytest-asyncio >=0.18.3, <1.0',
          'pytest-cov >=3.0, <4',
          'pytest-mock >=3.7.0, <4']}

entry_points = \
{'console_scripts': ['deploy-me = '
                     'foodx_devops_tools.deploy_me_entry:flit_entry',
                     'file-maintainer = '
                     'foodx_devops_tools.file_maintainer_entry:flit_entry',
                     'foodx-release-flow = '
                     'foodx_devops_tools.release_flow_entry:flit_entry',
                     'puff = foodx_devops_tools.puff_utility:entrypoint',
                     'validate-configuration = '
                     'foodx_devops_tools.validate_configuration:flit_entry']}

setup(name='foodx_devops_tools',
      version='0.15.1.post2',
      description='Foodx DevOps pipeline utilities.',
      author='FoodX Technologies',
      author_email='support@foodxtech.com',
      url='https://github.com/Food-X-Technologies/foodx_devops_tools',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.8',
     )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

from foodx_devops_tools.pipeline_config._structure import FrameFile


def test_clean():
    under_test = FrameFile(dir=pathlib.Path("some"), file=pathlib.Path("file"))

    assert under_test.path == pathlib.Path("some/file")


def test_none_dir_file():
    under_test = FrameFile(dir=None, file=None)

    assert under_test.path is None


def test_none_dir():
    under_test = FrameFile(dir=None, file=pathlib.Path("file"))

    assert under_test.path == pathlib.Path("file")


def test_none_file():
    under_test = FrameFile(dir=pathlib.Path("some"), file=None)

    assert under_test.path == pathlib.Path("some")



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import copy
import logging
import typing

import click

from foodx_devops_tools.azure.cloud.auth import (
    AzureAuthenticationError,
    login_service_principal,
)
from foodx_devops_tools.pipeline_config import (
    ApplicationDefinition,
    ApplicationDeploymentSteps,
    FlattenedDeployment,
    PipelineConfiguration,
    SingularFrameDefinition,
)
from foodx_devops_tools.pipeline_config.frames import (
    ApplicationStepDelay,
    ApplicationStepDeploymentDefinition,
    ApplicationStepScript,
)
from foodx_devops_tools.profiling import timing
from foodx_devops_tools.puff import PuffError

from ._dependency_monitor import wait_for_dependencies
from ._exceptions import DeploymentError
from ._state import PipelineCliOptions
from ._status import DeploymentState, DeploymentStatus, all_success
from .application_steps import delay_step, deploy_step, script_step

log = logging.getLogger(__name__)

STATUS_SLEEP_SECONDS = 1


def any_failed(values: typing.List[DeploymentState]) -> bool:
    """Evaluate if any deployment states failed."""
    result = any([x.code == DeploymentState.ResultType.failed for x in values])
    return result


def all_cancelled(values: typing.List[DeploymentState]) -> bool:
    """Evaluate if all deployments were cancelled."""
    result = all(
        [x.code == DeploymentState.ResultType.cancelled for x in values]
    )
    return result


async def assess_results(
    results: typing.List[DeploymentState],
) -> DeploymentState:
    """
    Condense an array of deployment results into a single result.

    Args:
        results: Array of results to condense.

    Returns:
        Success, if all results are success. Failed otherwise.
    """
    if all_success(results):
        log.debug("assessed result success")
        this_result = DeploymentState(code=DeploymentState.ResultType.success)
    elif any_failed(results):
        log.debug("assessed result failed")
        this_result = DeploymentState(code=DeploymentState.ResultType.failed)
    elif all_cancelled(results):
        log.debug("assessed result cancelled")
        this_result = DeploymentState(code=DeploymentState.ResultType.cancelled)
    else:
        # fallback is failure.
        log.debug("fallback to failed")
        messages = [
            x.message
            for x in results
            if x.message and (x.code != DeploymentState.ResultType.success)
        ]
        this_result = DeploymentState(
            code=DeploymentState.ResultType.failed, message=str(messages)
        )

    return this_result


async def _do_application_deployment(
    application_data: ApplicationDeploymentSteps,
    deployment_data: FlattenedDeployment,
    application_status: DeploymentStatus,
    enable_validation: bool,
) -> None:
    this_context = str(deployment_data.data.iteration_context)
    try:

        for this_step in application_data:
            with timing(log, this_context):
                if isinstance(this_step, ApplicationStepDeploymentDefinition):
                    await deploy_step(
                        this_step,
                        deployment_data,
                        enable_validation,
                    )
                elif isinstance(this_step, ApplicationStepScript):
                    await script_step(
                        this_step,
                        deployment_data,
                    )
                elif isinstance(this_step, ApplicationStepDelay):
                    await delay_step(this_step.delay_seconds)
                else:
                    raise DeploymentError(
                        "Bad application step definition, "
                        "{0}".format(this_context)
                    )

        log.info("application deployment succeeded, {0}".format(this_context))
        await application_status.write(
            this_context, DeploymentState.ResultType.success
        )
    except asyncio.CancelledError:
        message = "async cancelled signal"
        log.error(message)
        await application_status.write(
            this_context,
            DeploymentState.ResultType.cancelled,
            message,
        )
        raise
    except (AzureAuthenticationError, PuffError) as e:
        message = (
            "application deployment authentication "
            "failure, {0}, {1}, {2}".format(this_context, type(e), str(e))
        )
        await application_status.write(
            this_context,
            DeploymentState.ResultType.failed,
            message,
        )
        log.error(message)


async def deploy_application(
    application_data: ApplicationDefinition,
    deployment_data: FlattenedDeployment,
    application_status: DeploymentStatus,
    enable_validation: bool,
) -> None:
    """
    Deploy the steps of a frame application.

    Application steps must be deployed in sequence (serially).
    """
    this_context = str(deployment_data.data.iteration_context)
    try:
        message = "starting application deployment, {0}".format(this_context)
        log.info(message)
        click.echo(message)
        await application_status.initialize(this_context)

        await wait_for_dependencies(
            deployment_data.data.iteration_context,
            application_data.depends_on
            if application_data.depends_on
            else list(),
            application_status,
        )

        await application_status.write(
            this_context, DeploymentState.ResultType.in_progress
        )

        deploy_to = deployment_data.data.to
        application_name = deployment_data.context.application_name
        if deploy_to.application and (
            application_name != deploy_to.application
        ):
            await application_status.write(
                this_context,
                DeploymentState.ResultType.skipped,
                message="deployment targeted application, {0}".format(
                    str(deploy_to)
                ),
            )
        else:
            with timing(log, this_context):
                await _do_application_deployment(
                    application_data.steps,
                    deployment_data,
                    application_status,
                    enable_validation,
                )
    except Exception as e:
        message = "application deployment failed, {0}, {1}, {2}".format(
            this_context, type(e), str(e)
        )
        await application_status.write(
            this_context,
            DeploymentState.ResultType.failed,
            message,
        )
        log.exception(message)

    message = "application deployment completed, {0}".format(this_context)
    log.info(message)
    click.echo(click.style(message, fg="yellow"))


async def _do_frame_deployment(
    this_context: str,
    pipeline_parameters: PipelineCliOptions,
    deployment_data: FlattenedDeployment,
    frame_data: SingularFrameDefinition,
    frame_status: DeploymentStatus,
) -> None:
    # application status will show as "pending" until deployment activates.
    application_status = DeploymentStatus(
        this_context, pipeline_parameters.wait_timeout_seconds
    )
    try:
        wait_task = asyncio.create_task(
            application_status.wait_for_all_completed()
        )
        application_status.start_monitor()

        await wait_for_dependencies(
            deployment_data.data.iteration_context,
            frame_data.depends_on if frame_data.depends_on else list(),
            frame_status,
        )

        frame_deployment = copy.deepcopy(deployment_data)
        frame_deployment.data.frame_folder = frame_data.folder

        await asyncio.gather(
            *[
                deploy_application(
                    application_data,
                    frame_deployment.copy_add_application(application_name),
                    application_status,
                    pipeline_parameters.enable_validation,
                )
                for application_name, application_data in frame_data.applications.items()  # noqa: E501
            ],
            return_exceptions=False,
        )

        await wait_task

        message = "frame deployment completed, {0}".format(this_context)
        log.info(message)
        click.echo(click.style(message, fg="yellow"))
    except asyncio.TimeoutError:
        message = "timeout waiting for application deployment, {0}".format(
            deployment_data.data.iteration_context
        )
        log.error(message)
        click.echo(message, err=True)
        raise

    # read the state of the completed _applications_
    results = [
        await application_status.read(x)
        for x in await application_status.names()
    ]
    condensed_result = await assess_results(results)
    await frame_status.write(
        this_context, condensed_result.code, condensed_result.message
    )
    message = "frame deployment {0}, {1}".format(
        condensed_result.code.name, this_context
    )
    log.info(message)
    click.echo(
        click.style(
            message, fg=DeploymentStatus.STATE_COLOURS[condensed_result.code]
        )
    )


async def deploy_frame(
    frame_data: SingularFrameDefinition,
    deployment_data: FlattenedDeployment,
    frame_status: DeploymentStatus,
    pipeline_parameters: PipelineCliOptions,
) -> None:
    """
    Deploy the applications of a frame.

    Frame applications are deployed concurrently (in parallel).

    Raises:
        DeploymentTerminatedError: If any dependencies fail preventing
                                  completion.
    """
    this_context = str(deployment_data.data.iteration_context)
    message = "starting frame deployment, {0}".format(this_context)
    log.info(message)
    click.echo(message)

    await frame_status.initialize(this_context)
    await frame_status.write(
        this_context, DeploymentState.ResultType.in_progress
    )
    deploy_to = deployment_data.data.to
    frame_name = deployment_data.context.frame_name
    if deploy_to.frame and (frame_name != deploy_to.frame):
        await frame_status.write(
            this_context,
            DeploymentState.ResultType.skipped,
            message="deployment targeted frame, {0}".format(str(deploy_to)),
        )
    else:
        with timing(log, this_context):
            await _do_frame_deployment(
                this_context,
                pipeline_parameters,
                deployment_data,
                frame_data,
                frame_status,
            )


async def do_deploy(
    configuration: PipelineConfiguration,
    deployment_data: FlattenedDeployment,
    pipeline_parameters: PipelineCliOptions,
) -> DeploymentState:
    """
    Deploy the frames in a flattened deployment configuration.

    Raises:
        asyncio.TimeoutError: If the maximum wait time in
                              ``pipeline_parameters`` is exceeded.
    """
    this_frames = configuration.frames
    this_context = deployment_data.data.iteration_context.copy()
    deployment_data.data.iteration_context.append(
        deployment_data.data.deployment_tuple
    )

    frame_deployment_status = DeploymentStatus(
        str(deployment_data.data.iteration_context),
        timeout_seconds=pipeline_parameters.wait_timeout_seconds,
    )
    try:
        with timing(log, str(this_context)):
            # this is a temporary work-around to the login concurrency problem -
            # it works provided there is no more than a single subscription per
            # deployment at a time.
            # https://github.com/Food-X-Technologies/foodx_devops_tools/issues/129
            await login_service_principal(
                deployment_data.data.azure_credentials
            )

            wait_task = asyncio.create_task(
                frame_deployment_status.wait_for_all_completed()
            )
            frame_deployment_status.start_monitor()

            await asyncio.gather(
                *[
                    deploy_frame(
                        frame_data,
                        deployment_data.copy_add_frame(frame_name),
                        frame_deployment_status,
                        pipeline_parameters,
                    )
                    for frame_name, frame_data in this_frames.frames.items()
                ],
                return_exceptions=False,
            )
            await wait_task
    except asyncio.TimeoutError:
        message = "timeout waiting for frame deployments, {0}".format(
            deployment_data.data.iteration_context
        )
        log.error(message)
        click.echo(message, err=True)
        raise

    results = [
        await frame_deployment_status.read(x)
        for x in await frame_deployment_status.names()
    ]
    condensed_result = await assess_results(results)

    return condensed_result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import os
import pathlib
import re
import typing
from unittest.mock import AsyncMock

import pytest

from foodx_devops_tools.puff import run_puff
from foodx_devops_tools.puff.run import (
    DELETING_MESSAGE,
    GENERATING_MESSAGE,
    _acquire_yaml_filenames,
)
from tests.ci.support.capture import capture_stdout_stderr
from tests.ci.support.isolation import simple_isolated_filesystem


@contextlib.contextmanager
def mock_yaml_files(expected_files: typing.List[pathlib.Path]):
    with simple_isolated_filesystem() as dir:
        for this_file in [dir / x for x in expected_files]:
            os.makedirs(this_file.parent, exist_ok=True)
            with this_file.open(mode="w") as f:
                f.write("---")

        yield dir


class TestAcquireYamlFilenames:
    def test_no_files_found(self):
        with simple_isolated_filesystem() as dir:
            result = _acquire_yaml_filenames(dir, list())

        assert result == set()

    def test_yml(self):
        relative_files = [
            pathlib.Path("some/dir/a.yml"),
        ]
        with mock_yaml_files(relative_files) as dir:
            result = _acquire_yaml_filenames(dir, list())

        assert result == {dir / x for x in relative_files}

    def test_yaml(self):
        relative_files = [
            pathlib.Path("some/dir/a.yaml"),
        ]
        with mock_yaml_files(relative_files) as dir:
            result = _acquire_yaml_filenames(dir, list())

        assert result == {dir / x for x in relative_files}

    def test_no_ignore_patterns(self):
        relative_files = [
            pathlib.Path("some/dir/a.yml"),
            pathlib.Path("some/b.yaml"),
        ]
        with mock_yaml_files(relative_files) as dir:
            result = _acquire_yaml_filenames(dir, list())

        assert result == {dir / x for x in relative_files}

    def test_ignore_pattern1(self):
        relative_files = [
            pathlib.Path("some/dir/a.yml"),
            pathlib.Path("some/b.yaml"),
        ]
        ignore_pattern = ["some/dir/*"]
        with mock_yaml_files(relative_files) as dir:
            result = _acquire_yaml_filenames(dir, ignore_pattern)

        assert result == {dir / x for x in relative_files[1::]}

    def test_ignore_pattern2(self):
        relative_files = [
            pathlib.Path("some/dir/a.yml"),
            pathlib.Path("some/b.yaml"),
        ]
        ignore_pattern = ["**/dir/*"]
        with mock_yaml_files(relative_files) as dir:
            result = _acquire_yaml_filenames(dir, ignore_pattern)

        assert result == {dir / x for x in relative_files[1::]}

    def test_ignore_pattern3(self):
        relative_files = [
            pathlib.Path("some/dir/a.yml"),
            pathlib.Path("some/dir/b.yaml"),
            pathlib.Path("some/b.yaml"),
        ]
        ignore_pattern = ["**/b.yaml"]
        with mock_yaml_files(relative_files) as dir:
            result = _acquire_yaml_filenames(dir, ignore_pattern)

        assert result == {dir / x for x in relative_files[0:1]}

    def test_ignore_pattern4(self):
        relative_files = [
            pathlib.Path("some/a.yml"),
            pathlib.Path("some/dir/b.yaml"),
            pathlib.Path("some/deep/dir/b.yaml"),
            pathlib.Path("other/b.yaml"),
        ]
        ignore_pattern = ["some/**/*"]
        with mock_yaml_files(relative_files) as dir:
            result = _acquire_yaml_filenames(dir, ignore_pattern)

        assert result == {dir / x for x in relative_files[3:4]}


class TestRunPuff:
    @pytest.mark.asyncio
    async def test_generate_clean(self, capsys, mocker, mock_async_method):
        """Empty ignore patterns run acquires all YAML files."""
        mock_path = mocker.create_autospec(pathlib.Path)
        mock_path.is_dir.return_value = False
        mock_path.is_file.return_value = True
        mock_path.stem = "this_file"
        mock_path.parent = pathlib.Path(".")
        mocker.patch(
            "foodx_devops_tools.puff.run.load_puffignore",
            side_effect=AsyncMock(return_value=list()),
        )
        mock_async_method(
            "foodx_devops_tools.puff.arm.load_yaml",
            return_value={"environments": dict()},
        )
        mock_save = mock_async_method(
            "foodx_devops_tools.puff.arm._save_parameter_file"
        )

        await run_puff(mock_path, False, False)

        result = capsys.readouterr().out
        assert GENERATING_MESSAGE in result

        mock_save.assert_has_calls(
            [
                mocker.call(
                    pathlib.Path("this_file.json"), mocker.ANY, mocker.ANY
                )
            ]
        )

    @pytest.mark.asyncio
    async def test_output_dir_clean(self, capsys, mocker, mock_async_method):
        """Output directory specified for output."""
        mock_path = mocker.create_autospec(pathlib.Path)
        mock_path.is_dir.return_value = False
        mock_path.is_file.return_value = True
        mock_path.stem = "this_file"
        mock_path.parent = pathlib.Path(".")
        mocker.patch(
            "foodx_devops_tools.puff.run.load_puffignore",
            side_effect=AsyncMock(return_value=list()),
        )
        mock_async_method(
            "foodx_devops_tools.puff.arm.load_yaml",
            return_value={"environments": dict()},
        )
        mock_save = mock_async_method(
            "foodx_devops_tools.puff.arm._save_parameter_file"
        )

        await run_puff(
            mock_path, False, False, output_dir=pathlib.Path("some/dir")
        )

        result = capsys.readouterr().out
        assert GENERATING_MESSAGE in result

        mock_save.assert_has_calls(
            [
                mocker.call(
                    pathlib.Path("some/dir/this_file.json"),
                    mocker.ANY,
                    mocker.ANY,
                )
            ]
        )

    @pytest.mark.asyncio
    async def test_delete_clean(self, capsys, mocker, mock_async_method):
        """Empty ignore patterns run acquires all YAML files."""
        mock_path = mocker.create_autospec(pathlib.Path)
        mock_path.is_dir.return_value = False
        mock_path.is_file.return_value = True
        mock_path.stem.return_value = "this_file"
        mocker.patch(
            "foodx_devops_tools.puff.run.load_puffignore",
            side_effect=AsyncMock(return_value=list()),
        )
        mock_async_method(
            "foodx_devops_tools.puff.arm.load_yaml",
            return_value={"environments": dict()},
        )
        mock_async_method("foodx_devops_tools.puff.arm._delete_parameter_file")

        await run_puff(mock_path, True, False)

        result = capsys.readouterr().out
        assert DELETING_MESSAGE in result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Applications, frames configuration I/O."""

import copy
import enum
import logging
import pathlib
import typing

import pydantic

from ._exceptions import FrameDefinitionsError
from ._loader import load_configuration
from ._structure import FrameFile, StructuredName, StructuredPathCollection

log = logging.getLogger(__name__)

ENTITY_NAME = "frames"
ENTITY_SINGULAR = "frame"

DependencyDeclarations = typing.List[str]
GlobPathDeclarations = typing.List[str]


@enum.unique
class DeploymentMode(str, enum.Enum):
    """Azure Cloud resource group deployment mode enumeration."""

    complete = "Complete"
    incremental = "Incremental"


class TriggersDefinition(pydantic.BaseModel):
    """Definition of deployment triggers."""

    paths: GlobPathDeclarations


class ApplicationStepDelay(pydantic.BaseModel):
    """Define a pause delay for an application step."""

    delay_seconds: int


class ApplicationStepDeploymentDefinition(pydantic.BaseModel):
    """Application resource group deployment definition."""

    mode: DeploymentMode
    name: str
    resource_group: str

    arm_file: typing.Optional[pathlib.Path]
    puff_file: typing.Optional[pathlib.Path]
    static_secrets: typing.Optional[bool] = False


class ApplicationStepScript(pydantic.BaseModel):
    """Execute arbitrary shell commands during deployment."""

    name: str
    script: str


ApplicationDeploymentSteps = typing.List[
    typing.Union[
        ApplicationStepDelay,
        ApplicationStepDeploymentDefinition,
        ApplicationStepScript,
    ]
]


class ApplicationDefinition(pydantic.BaseModel):
    """A single application definition."""

    depends_on: typing.Optional[DependencyDeclarations]
    steps: ApplicationDeploymentSteps

    @pydantic.validator("steps")
    def check_applications(
        cls: pydantic.BaseModel, steps_candidate: ApplicationDeploymentSteps
    ) -> ApplicationDeploymentSteps:
        """Validate ``steps`` field."""
        step_names: typing.List[str] = [
            x.name
            for x in steps_candidate
            if isinstance(x, ApplicationStepDeploymentDefinition)
        ]
        if len(step_names) != len(set(step_names)):
            message = "Application step names must be unique, {0}".format(
                str(step_names)
            )
            # log the message here because pydantic exception handling
            # masks the true exception that caused a validation failure.
            log.error(message)
            raise ValueError(message)

        return steps_candidate


ApplicationDeclarations = typing.Dict[str, ApplicationDefinition]


class SingularFrameDefinition(pydantic.BaseModel):
    """A single frame definition."""

    applications: ApplicationDeclarations
    folder: pathlib.Path

    depends_on: typing.Optional[DependencyDeclarations]
    triggers: typing.Optional[TriggersDefinition]

    @pydantic.validator("applications")
    def check_applications(
        cls: pydantic.BaseModel, applications_candidate: dict
    ) -> dict:
        """Validate ``applications`` field."""
        application_names: list = list(applications_candidate.keys())
        if any([not x for x in application_names]):
            message = "Empty {0} names prohibited".format(ENTITY_SINGULAR)
            # log the message here because pydantic exception handling
            # masks the true exception that caused a validation failure.
            log.error(message)
            raise ValueError(message)
        if len(set(application_names)) != len(application_names):
            message = "Duplicate {0} names prohibited".format(ENTITY_SINGULAR)
            # log the message here because pydantic exception handling
            # masks the true exception that caused a validation failure.
            log.error(message)
            raise ValueError(message)

        return applications_candidate

    @pydantic.root_validator
    def check_dependencies(
        cls: pydantic.BaseModel, applications_candidate: dict
    ) -> dict:
        """
        Validate frame dependencies.

        Frames must only declare dependencies on other frames.
        """
        applications: dict = applications_candidate.get("applications", dict())
        frame_names: set = set(applications.keys())
        for name, item in applications.items():
            if item.depends_on and (
                any([x not in frame_names for x in item.depends_on])
            ):
                message = "Unknown application dependency, {0}".format(
                    str(item.depends_on)
                )
                # log the message here because pydantic exception handling
                # masks the true exception that caused a validation failure.
                log.error(message)
                raise ValueError(message)

        return applications_candidate


FrameDeclarations = typing.Dict[str, SingularFrameDefinition]

U = typing.TypeVar("U", bound="FramesTriggersDefinition")


class FramesTriggersDefinition(pydantic.BaseModel):
    """Definition of frames and triggers."""

    frames: FrameDeclarations
    triggers: typing.Optional[TriggersDefinition]

    @pydantic.validator(ENTITY_NAME)
    def check_frames(cls: pydantic.BaseModel, frames_candidate: dict) -> dict:
        """Validate ``frames`` field."""
        frame_names: list = list(frames_candidate.keys())
        if any([not x for x in frame_names]):
            message = "Empty {0} names prohibited".format(ENTITY_SINGULAR)
            # log the message here because pydantic exception handling
            # masks the true exception that caused a validation failure.
            log.error(message)
            raise ValueError(message)
        if len(set(frame_names)) != len(frame_names):
            message = "Duplicate {0} names prohibited".format(ENTITY_SINGULAR)
            # log the message here because pydantic exception handling
            # masks the true exception that caused a validation failure.
            log.error(message)
            raise ValueError(message)

        return frames_candidate

    @pydantic.root_validator
    def check_dependencies(
        cls: pydantic.BaseModel, frames_candidate: dict
    ) -> dict:
        """
        Validate frame dependencies.

        Frames must only declare dependencies on other frames.
        """
        frames: dict = frames_candidate.get("frames", dict())
        frame_names: set = set(frames.keys())
        for name, item in frames.items():
            if item.depends_on and (
                any([x not in frame_names for x in item.depends_on])
            ):
                message = "Unknown frame dependency, {0}".format(
                    str(item.depends_on)
                )
                # log the message here because pydantic exception handling
                # masks the true exception that caused a validation failure.
                log.error(message)
                raise ValueError(message)

        return frames_candidate

    def arm_file_paths(self: U) -> StructuredPathCollection:
        """
        Generate collection of ARM template file paths.

        Paths not specified are defined as ``None``. Default path inferences
        are calculated elsewhere in the application.

        Returns:
            Collection of ARM template file paths indexed by structured name.
        """
        result: StructuredPathCollection = dict()
        for frame_name, frame_data in self.frames.items():
            frame_structure = StructuredName()
            frame_structure.append(frame_name)

            this_folder = frame_data.folder
            for (
                application_name,
                application_data,
            ) in frame_data.applications.items():
                app_structure = copy.deepcopy(frame_structure)
                app_structure.append(application_name)
                for this_step in application_data.steps:
                    if isinstance(
                        this_step, ApplicationStepDeploymentDefinition
                    ):
                        step_structure = copy.deepcopy(app_structure)
                        step_structure.append(this_step.name)

                        arm_name: typing.Optional[pathlib.Path]
                        if this_step.arm_file:
                            arm_name = this_step.arm_file
                        else:
                            # default names are inferred elsewhere
                            arm_name = None

                        result[step_structure] = FrameFile(
                            dir=this_folder, file=arm_name
                        )
        return result

    def puff_file_paths(self: U) -> StructuredPathCollection:
        """
        Generate collection of source puff file paths.

        Paths not specified are defined as ``None``. Default path inferences
        are calculated elsewhere in the application.

        Returns:
            Collection of YAML puff file paths indexed by structured name.
        """
        result: StructuredPathCollection = dict()
        for frame_name, frame_data in self.frames.items():
            frame_structure = StructuredName()
            frame_structure.append(frame_name)

            this_folder = frame_data.folder
            for (
                application_name,
                application_data,
            ) in frame_data.applications.items():
                app_structure = copy.deepcopy(frame_structure)
                app_structure.append(application_name)
                for this_step in application_data.steps:
                    if isinstance(
                        this_step, ApplicationStepDeploymentDefinition
                    ):
                        step_structure = copy.deepcopy(app_structure)
                        step_structure.append(this_step.name)

                        puff_name: typing.Optional[pathlib.Path]
                        if this_step.puff_file:
                            puff_name = this_step.puff_file
                        else:
                            # default names are inferred elsewhere
                            puff_name = None

                        result[step_structure] = FrameFile(
                            dir=this_folder, file=puff_name
                        )
        return result

    def frame_folders(self: U) -> StructuredPathCollection:
        """
        Generate collection of frame folder paths.

        Returns:
            Collection of frame folder paths indexed by structured name.
        """
        result: StructuredPathCollection = dict()
        for frame_name, frame_data in self.frames.items():
            this_structure = StructuredName()
            this_structure.append(frame_name)
            result[this_structure] = FrameFile(
                dir=frame_data.folder, file=pathlib.Path("")
            )

        return result


ValueType = FramesTriggersDefinition

T = typing.TypeVar("T", bound="FramesDefinition")


class FramesDefinition(pydantic.BaseModel):
    """Definition of frames."""

    frames: ValueType


def load_frames(file_path: pathlib.Path) -> FramesDefinition:
    """
    Load frame definitions from file.

    Args:
        file_path: Path to frame definitions file.

    Returns:
        Frame definitions.
    Raises:
        FrameDefinitionsError: If an error occurs loading the file.
    """
    log.info("loading frame definitions from file, {0}".format(file_path))
    result = load_configuration(
        file_path, FramesDefinition, FrameDefinitionsError, ENTITY_NAME
    )

    return typing.cast(FramesDefinition, result)
